
#include <WiFi.h>
#include <ESPAsyncWebServer.h>
#include <AsyncTCP.h>
#include <Hash.h>
#include "FS.h"
#include "SPIFFS.h"

#define FORMAT_SPIFFS_IF_FAILED true

const char ledPin = 2;

String httpUsername = "admin";
String httpPassword = "admin";

String _AP_ssid = "Esp_AP";
String _AP_password = "12345678";
IPAddress local_IP(192,168,7,22);
IPAddress gateway(192,168,7,1);
IPAddress subnet(255,255,255,0);

AsyncWebServer server(80);

bool wifiIsConnected = false;

String ledValue ="false";
String ledState ="The LED is off!";

const char* param_ledvalue = "ledvalue";
const char* param_ap_ssid = "inputApSsid";
const char* param_ap_password = "inputApPassword";
const char* param_wifi_ssid = "inputWifiSsid";
const char* param_wifi_password = "inputWifiPassword";
const char* param_httpUsername = "inputHttpUsername";
const char* param_httpPassword = "inputHttpPassword";


const char index_html[] PROGMEM = R"rawliteral(
<!DOCTYPE HTML>
<html>
  <head>
    <title>Home page</title>
    <meta charset="UTF-8">
    <meta name="viewport" content="width=device-width, initial-scale=1">
    <script>
      function submitData() 
      {
        setTimeout(function(){
          document.location.reload(false); 
          }, 1000);   
      }
    </script>
  </head>
  <body>
    <center>
      <h2>ESP32 Asynchronous web server</h2>
      <div style="height:50px"></div>

      <p>%LED_STATE%</p>

      <form action="/ledSwitch" target="self_page">
        <input type="hidden" name="ledvalue" value="%LED_VALUE%">
        <input type="submit" value="LED" onclick="submitData()">
      </form>

      <br><br>
      <button onclick="window.location.href='/settings';">Settings</button>
      <iframe style="display:none" name="self_page"></iframe>
    </center>
  </body>
</html> )rawliteral";


const char settings_html[] PROGMEM = R"rawliteral(
<!DOCTYPE HTML>
<html>
  <head>
    <title>Settings</title>
    <meta charset="UTF-8">
    <meta name="viewport" content="width=device-width, initial-scale=1">
    <script>
      function submitData() 
      {
        setTimeout(function(){
          document.location.reload(false); 
          }, 1000);   
      }
      function logoutButton() 
      {
        var xmlHttpRequest = new XMLHttpRequest();
        xmlHttpRequest.open("GET", "/logout", true);
        xmlHttpRequest.send();
        setTimeout(function() {
          window.open("/","/"); 
          }, 1000);
      }
    </script>
  </head>
  <body>
    <center>
      <h2>Settings</h2>
      <div style="height:50px"></div>

      <fieldset style="width:300px">
        <legend>Network settings:</legend>
        <br><br>
        <form action="/network" target="self_page">

          AP SSID:<br>
          <input type="text" value='%inputApSsid%' name="inputApSsid" size="20">
          <br><br>
          
          AP Password:<br>
          <input type="text" value='%inputApPassword%' name="inputApPassword" size="20">
          <br>( at least 8 characters! )
          <br><br>
        
          Wifi SSID:<br>
          <input type="text" value='%inputWifiSsid%' name="inputWifiSsid" size="20">
          <br><br>
          
          Wifi Password:<br>
          <input type="text" value='%inputWifiPassword%' name="inputWifiPassword" size="20">
          <br><br>
          
          <input type="submit" value="Save" onclick="submitData()">
        </form>
        <br><br>
      </fieldset>

      <div style="height:20px"></div>

      <fieldset style="width:300px">
        <legend>Login-Daten:</legend>
        <br><br>
        <form action="/http_login_set" target="self_page">
        
          User name:<br>
          <input type="text" value='%inputHttpUsername%'  name="inputHttpUsername" size="20">
          <br><br>
          
          Password:<br>
          <input type="text" value='%inputHttpPassword%'  name="inputHttpPassword" size="20">
          <br><br><br>
          
          <input type="submit" value="Save" onclick="submitData()">
        </form>
        <br><br>
      </fieldset>

      <div style="height:20px"></div>
      
      <fieldset style="width:300px">
        <legend>To home page</legend>
        <br><br>
        <button onclick="window.location.href='/';">Home page</button>
        <br><br>
      </fieldset>
      
      <div style="height:20px"></div>
      
      <fieldset style="width:300px">
        <legend>Logout</legend>
        <br><br>
        <button onclick="logoutButton()">Logout</button>
        <br><br>
      </fieldset>

      <div style="height:50px;"></div>
      <iframe style="display:none" name="self_page"></iframe>
    </center>
  </body>
</html> )rawliteral";

String inputValueAdapter(const String& inputValueSet)
{
  if(inputValueSet == "LED_STATE")
  {
    return ledState;
  }
  if(inputValueSet == "LED_VALUE")
  {
    return ledValue;
  }
  if(inputValueSet == "inputApSsid")
  {
    return readFile(SPIFFS, "/apSsid.txt");
  }

  if(inputValueSet == "inputApPassword")
  {
    return readFile(SPIFFS, "/apPassword.txt");
  }

  if(inputValueSet == "inputWifiSsid")
  {
    return readFile(SPIFFS, "/wifiSsid.txt");
  }

  if(inputValueSet == "inputWifiPassword")
  {
    return readFile(SPIFFS, "/wifiPassword.txt");
  }
  
  if(inputValueSet == "inputHttpUsername")
  {
    return readFile(SPIFFS, "/httpUsername.txt");
  }
  
  if(inputValueSet == "inputHttpPassword")
  {
    return readFile(SPIFFS, "/httpPassword.txt");
  }
  return String();
}

void setupAsyncServer()
{
  server.on("/", HTTP_GET, [](AsyncWebServerRequest *request)
  {
    request->send_P(200, "text/html", index_html, inputValueAdapter);
  });

  server.on("/ledSwitch", HTTP_GET, [](AsyncWebServerRequest *request)
  {
    ledValue = request->getParam(param_ledvalue)->value();

    if(ledValue == "true")
    {
      ledValue = "false";
      digitalWrite(ledPin, LOW);
      ledState =" The LED is off!";
    }
    else if(ledValue == "false")
    {
      ledValue = "true";
      digitalWrite(ledPin, HIGH);
      ledState ="The LED is on!";
    }
    request->send(200, "text/text", ledValue);
  });

  server.on("/settings", HTTP_GET, [](AsyncWebServerRequest *request)
  {
    String temp_httpUsername = readFile(SPIFFS, "/httpUsername.txt");
    if(temp_httpUsername == "")
    {
      temp_httpUsername = httpUsername;
    }
    const char* http_username = temp_httpUsername.c_str();
    String temp_httpPassword = readFile(SPIFFS, "/httpPassword.txt");
    if(temp_httpPassword == "")
    {
      temp_httpPassword = httpPassword;
    }
    const char* http_password = temp_httpPassword.c_str();
    
    if(!request->authenticate(http_username, http_password))
    {
      return request->requestAuthentication();
    }

    request->send_P(200, "text/html", settings_html, inputValueAdapter);
  });

  server.on("/network", HTTP_GET, [](AsyncWebServerRequest *request)
  {
    String temp_httpUsername = readFile(SPIFFS, "/httpUsername.txt");
    if(temp_httpUsername == "")
    {
      temp_httpUsername = httpUsername;
    }
    const char* http_username = temp_httpUsername.c_str();

    String temp_httpPassword = readFile(SPIFFS, "/httpPassword.txt");
    if(temp_httpPassword == "")
    {
      temp_httpPassword = httpPassword;
    }
    const char* http_password = temp_httpPassword.c_str();
    
    if(!request->authenticate(http_username, http_password))
    {
      return request->requestAuthentication();
    }

    String inputMessage;
    if (request->hasParam(param_ap_ssid)) 
    {
      inputMessage = request->getParam(param_ap_ssid)->value();
      writeFile(SPIFFS, "/apSsid.txt", inputMessage.c_str());
    }
                                                                          
    if (request->hasParam(param_ap_password))
    {
      inputMessage = request->getParam(param_ap_password)->value();
      writeFile(SPIFFS, "/apPassword.txt", inputMessage.c_str());
    }

    if (request->hasParam(param_wifi_ssid)) 
    {
      inputMessage = request->getParam(param_wifi_ssid)->value();
      writeFile(SPIFFS, "/wifiSsid.txt", inputMessage.c_str());
    }     

    if (request->hasParam(param_wifi_password)) 
    {
      inputMessage = request->getParam(param_wifi_password)->value();
      writeFile(SPIFFS, "/wifiPassword.txt", inputMessage.c_str());
    }

    request->send(200, "text/text", inputMessage);
    
    Serial.println("ESP Restart...");
    ESP.restart();
  });

  server.on("/http_login_set", HTTP_GET, [](AsyncWebServerRequest *request)
  {
    String temp_httpUsername = readFile(SPIFFS, "/httpUsername.txt");
    if(temp_httpUsername == "")
    {
      temp_httpUsername = httpUsername;
    }
    const char* http_username = temp_httpUsername.c_str();

    String temp_httpPassword = readFile(SPIFFS, "/httpPassword.txt");
    if(temp_httpPassword == "")
    {
      temp_httpPassword = httpPassword;
    }
    const char* http_password = temp_httpPassword.c_str();
    
    if(!request->authenticate(http_username, http_password))
    {
      return request->requestAuthentication();
    }

    String inputMessage;                                
    if (request->hasParam(param_httpUsername))
    {
      inputMessage = request->getParam(param_httpUsername)->value();
      writeFile(SPIFFS, "/httpUsername.txt", inputMessage.c_str());
    }

    if (request->hasParam(param_httpPassword))
    {
      inputMessage = request->getParam(param_httpPassword)->value();
      writeFile(SPIFFS, "/httpPassword.txt", inputMessage.c_str());
    }
    request->send(200, "text/text", inputMessage);
  });
  
  server.on("/logout", HTTP_GET, [](AsyncWebServerRequest *request)
  {
    request->send(401);
  });

  server.onNotFound(notFound);
  server.begin();
}

void notFound(AsyncWebServerRequest *request) 
{
  request->send(404, "text/plain", "The page not found");
}

String readFile(fs::FS &fs, const char * path)
{
  String fileContent = "";
  File file = fs.open(path, "r");
  if(!file || file.isDirectory())
  {
    return fileContent;
  }
  while(file.available())
  {
    fileContent+=String((char)file.read());
  }
  file.close();
  return fileContent;
}

void writeFile(fs::FS &fs, const char * path, const char * message)
{
  File file = fs.open(path, "w");
  if(!file)
  {
    return;
  }
  file.print(message);
  file.close();
}

void setup_ap()
{
  String temp_pass = readFile(SPIFFS, "/apPassword.txt");
  if(temp_pass == "")
  {
    temp_pass = _AP_password;
  }
  String temp_ssid = readFile(SPIFFS, "/apSsid.txt");
  if(temp_ssid == "")
  {
    temp_ssid = _AP_ssid;
  }

  const char* AP_ssid = temp_ssid.c_str();
  const char* AP_password = temp_pass.c_str();
  
  WiFi.softAPConfig(local_IP, gateway, subnet);
  WiFi.softAP(AP_ssid, AP_password);
  IPAddress Ap_Ip = WiFi.softAPIP();
  
  Serial.print("Access Point IP address: ");
  Serial.println(Ap_Ip);
  Serial.println();
}

void setup_wifi() 
{
  String temp_pass = readFile(SPIFFS, "/wifiPassword.txt");
  char passbuff[temp_pass.length() + 1];
  temp_pass.toCharArray(passbuff, temp_pass.length() + 1);
  const char* password = passbuff;

  String temp_ssid = readFile(SPIFFS, "/wifiSsid.txt");
  char ssidbuff[temp_ssid.length() + 1];
  temp_ssid.toCharArray(ssidbuff, temp_ssid.length() + 1);
  const char* ssid = ssidbuff;

  WiFi.begin(ssid, password);
  Serial.print("Connection to WiFi network.");
  int i = 0;
  while (true) 
  {
    if(WiFi.status() != WL_CONNECTED)
    {
      delay(500);
      Serial.print(".");
      i++;
      if(i > 30)
      {
        wifiIsConnected = false;
        Serial.println();
        Serial.println("It cannot connect to the WiFi network.");
        break;
      }
    }
    else
    {
      wifiIsConnected = true;
      Serial.println();
      Serial.println("Connected!");
      Serial.print("Ip address: ");
      Serial.println(WiFi.localIP());
      Serial.println();
      break;
    }
  }
}

void setup()
{
	pinMode(ledPin, OUTPUT);
  digitalWrite(ledPin, HIGH);
  delay(2000);
  digitalWrite(ledPin, LOW);
  
  Serial.begin(115200);
  SPIFFS.begin();

  WiFi.mode(WIFI_STA);
  setup_wifi();

  if(!wifiIsConnected)
  {
    WiFi.mode(WIFI_AP);
    setup_ap();    
  }
  
  setupAsyncServer();
}

void loop(){}



